#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : routing.py
@Author : yell
@Desc : routing framenwork
'''
import os
from engine.base import EngineBase
from database.enum import EdaTool, FlowStep
from flow.flow_db import DbFlow
from tools.iEDA.module.routing import IEDARouting
from tools.innovus.module.routing import InnovusRouting
if os.environ.get('OpenROAD') == "on":
    from tools.OpenROAD.module.routing import OpenROADRouting

class EngineRouting(EngineBase):
    """routing framenwork"""       
    def run(self):
        if(self.step.step == FlowStep.route):
            return self.run_route()
        if(self.step.step == FlowStep.globalRouting):
            return self.run_global_routing()
        if(self.step.step == FlowStep.detailRouting):
            return self.run_detail_routing()

    #######################################################################################    
    # route
    #######################################################################################
    def run_route(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_ieda_route()
        elif(self.eda_tool == EdaTool.INNOVUS):
            return self.run_innovus_route()
        elif(self.eda_tool == EdaTool.OPENROAD):
            if os.environ.get('OpenROAD') == "on":
                return self.run_openroad_route()
            else:
                print("Error, do not support OpenROAD, please install OpenROAD and set OpenROAD environment to 'on'")
                return None
        else:
            print("Error, don't support this tool : " + self.eda_tool.value)
            exit(0)
        
    def run_ieda_route(self):
        # build flow
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.legalization)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                        pre_flow=self.pre_step,
                        step=FlowStep.route)
        self.build_path()
        
        # run iEDA
        run_flow = IEDARouting(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_routing()
        
        feature_result = self.run_feature_ieda()
        
        run_flow.close_routing()
        
        return feature_result
    
    def run_innovus_route(self):
         # build flow
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.INNOVUS,
                              step=FlowStep.cts)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.INNOVUS,
                          pre_flow=self.pre_step,
                          step=FlowStep.route)
            
        self.build_path()
        
        # run innovus
        run_flow = InnovusRouting(dir_workspace = self.dir_workspace,
                               dir_resource = self.dir_resource,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step,
                              task = self.task)
        run_flow.run_routing()
        
        # update result
        # self.result.output_features['summary_routing'] = feature_path
        
        return self.get_result()
    
    # route divide into two steps : gr & dr, so the io files need to adjust to the changes
    def run_openroad_route(self):
        # store output 
        if self.output_def == None:
            output_def = self.workspace.get_output_def(self.step)
        else:
            output_def = self.output_def
        if self.output_verilog == None:
            output_verilog = self.workspace.get_output_verilog(self.step)
        else:
            output_verilog = self.output_verilog
        # step 1 run gr
        # reset step
        if(self.step == None):
            print("Please set step")
            exit(0)
            
        self.step.step = FlowStep.globalRouting
        # change output to gr flow
        self.output_def = self.workspace.get_output_def(self.step)
        self.output_verilog = self.workspace.get_output_verilog(self.step)
        
        self.run_openroad_gr()
        # run dr
        # reset prestep & step
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.OPENROAD,
                              step=FlowStep.globalRouting)
        else:
            self.pre_step.step=FlowStep.globalRouting
            
        self.step = DbFlow(eda_tool=EdaTool.OPENROAD,
                          pre_flow=self.pre_step,
                          step=FlowStep.detailRouting)
        # change input to gr 
        self.input_def = self.workspace.get_output_def(self.pre_step)
        self.input_verilog = self.workspace.get_output_verilog(self.pre_step)
        # restore output to origin setting
        self.output_def = output_def
        self.output_verilog = output_verilog
        
        self.run_openroad_dr()
        
        return self.get_result()
    
    #######################################################################################    
    # global routing
    #######################################################################################
    def run_global_routing(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_ieda_gr()
        elif(self.eda_tool == EdaTool.INNOVUS):
            return self.run_innovus_gr()
        elif(self.eda_tool == EdaTool.OPENROAD):
            if os.environ.get('OpenROAD') == "on":
                return self.run_openroad_gr()
            else:
                print("Error, do not support OpenROAD, please install OpenROAD and set OpenROAD environment to 'on'")
                return None
        else:
            print("Error, don't support this tool : " + self.eda_tool.value, + ", step : " + self.step.step.value)
            exit(0)
    
    def run_ieda_gr(self):
        # build flow
        print("Error, don't support this tool : " + self.eda_tool.value, + ", step : " + self.step.step.value)
        pass
    
    def run_innovus_gr(self):
         # build flow
        print("Error, don't support this tool : " + self.eda_tool.value, + ", step : " + self.step.step.value)
        pass
    
    def run_openroad_gr(self):
         # build flow
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.OPENROAD,
                              step=FlowStep.legalization)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.OPENROAD,
                          pre_flow=self.pre_step,
                          step=FlowStep.globalRouting)
            
        self.build_path()
        
        # run openroad
        run_flow = OpenROADRouting(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step,
                              task = self.task)
        run_flow.run_global_routing()
        
        return self.get_result()
    
#######################################################################################    
    # detail routing
    #######################################################################################
    def run_detail_routing(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_ieda_dr()
        elif(self.eda_tool == EdaTool.INNOVUS):
            return self.run_innovus_dr()
        elif(self.eda_tool == EdaTool.OPENROAD):
            if os.environ.get('OpenROAD') == "on":
                return self.run_openroad_dr()
            else:
                print("Error, do not support OpenROAD, please install OpenROAD and set OpenROAD environment to 'on'")
                return None
        else:
            print("Error, don't support this tool : " + self.eda_tool.value, + ", step : " + self.step.step.value)
            exit(0)
    
    def run_ieda_dr(self):
        # build flow
        print("Error, don't support this tool : " + self.eda_tool.value, + ", step : " + self.step.step.value)
        pass
    
    def run_innovus_dr(self):
         # build flow
        print("Error, don't support this tool : " + self.eda_tool.value, + ", step : " + self.step.step.value)
        pass
    
    def run_openroad_dr(self):
         # build flow
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.OPENROAD,
                              step=FlowStep.globalRouting)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.OPENROAD,
                          pre_flow=self.pre_step,
                          step=FlowStep.detailRouting)
            
        self.build_path()
        
        # run openroad
        run_flow = OpenROADRouting(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step,
                              task = self.task)
        run_flow.run_detail_routing()
        
        return self.get_result()


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : placement.py
@Author : yell
@Desc : placement framework
'''
import os

from engine.base import EngineBase
from database.enum import EdaTool, FlowStep
from flow.flow_db import DbFlow
from tools.iEDA.module.placement import IEDAPlacement
from tools.innovus.module.placement import InnovusPlacement

if os.environ.get('DREAMPlace') == "on":
    from tools.DREAMPlace.run_dreamplace import DreamPlace
if os.environ.get('Xplace') == "on":
    from tools.Xplace.run_xplace import RunXPlace
if os.environ.get('OpenROAD') == "on":
    from tools.OpenROAD.module.placement import OpenROADPlacement

class EnginePlacement(EngineBase):
    """placement framework"""
    def run(self):
        if(self.step.step == FlowStep.prePlace):
            return self.run_pre_placer()
        if(self.step.step == FlowStep.place):
            return self.run_placer()
        if(self.step.step == FlowStep.globalPlace):
            return self.run_global_placement()
        if(self.step.step == FlowStep.detailPlace):
            return self.run_detail_placement()
        if(self.step.step == FlowStep.legalization):
            return self.run_legalization()
        if(self.step.step == FlowStep.filler):
            return self.run_filler()

    #######################################################################################    
    # placement
    #######################################################################################
    def run_pre_placer(self):
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_pre_placer_innovus()
        else:
            print("only supper innovus pre place.")
            exit(0)
        
    
    def run_pre_placer_innovus(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.INNOVUS,
                              step=FlowStep.floorplan)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.INNOVUS,
                          pre_flow=self.pre_step,
                          step=FlowStep.prePlace)
            
        self.build_path()
        
        # run iEDA
        run_flow = InnovusPlacement(dir_workspace = self.dir_workspace,
                               dir_resource = self.dir_resource,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step,
                              task = self.task)
        run_flow.run_pre_place()
        
        # update result
        # self.result.output_features['summary_preplace'] = feature_path
        
        return self.get_result()

    #######################################################################################    
    # placement
    #######################################################################################
    def run_placer(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_placer_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_placer_innovus()
        
        if(self.eda_tool == EdaTool.DREAMPLACE):
            if os.environ.get('DREAMPlace') == "on":
                return self.run_placer_dreamplace()
            else:
                print("Error, do not support DREAMPlace, please install DREAMPlace and set DREAMPlace environment to 'on'")
                return None
        
        if(self.eda_tool == EdaTool.XPLACE):
            if os.environ.get('Xplace') == "on":
                return self.run_placer_xplace()
            else:
                print("Error, do not support XPlace, please install XPlace and set Xplace environment to 'on'")
                return None
        
    def run_placer_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.fixFanout)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                        pre_flow=self.pre_step,
                        step=FlowStep.place)
        self.build_path()
        
        # run iEDA
        run_flow = IEDAPlacement(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_placement()
        
        return self.run_feature_ieda()
    
    def run_placer_innovus(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.INNOVUS,
                              step=FlowStep.prePlace)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.INNOVUS,
                          pre_flow=self.pre_step,
                          step=FlowStep.place)
            
        self.build_path()
        
        # run innovus
        run_flow = InnovusPlacement(dir_workspace = self.dir_workspace,
                               dir_resource = self.dir_resource,
                              input_def = self.input_def, 
                              input_verilog = self.input_verilog,
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step,
                              task = self.task)
        run_flow.run_place()
        
        # update result
        # self.result.output_features['summary_placement'] = feature_path
        
        return self.get_result()
    
    def run_placer_dreamplace(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.fixFanout)
        self.step = DbFlow(eda_tool=EdaTool.DREAMPLACE,
                        pre_flow=self.pre_step,
                        step=FlowStep.place)
        self.build_path()
        
        # run DREAMPlace
        run_flow = DreamPlace(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              input_verilog= self.input_verilog, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_placement()
        
        return self.run_feature_ieda(True)
    
    def run_placer_xplace(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.fixFanout)
        self.step = DbFlow(eda_tool=EdaTool.XPLACE,
                        pre_flow=self.pre_step,
                        step=FlowStep.place)
        self.build_path()
        
        # run xplace
        run_flow = RunXPlace(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              input_verilog= self.input_verilog, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_placement()
        
        return self.run_feature_ieda(True)
    
    #######################################################################################    
    # global placement
    #######################################################################################
    def run_global_placement(self):  
        if(self.eda_tool == EdaTool.OPENROAD):
            if os.environ.get('OpenROAD') == "on":
                return self.run_gp_OpenROAD()
            else:
                print("Error, do not support OpenROAD, please install OpenROAD and set OpenROAD environment to 'on'")
                return None
        else:
            print("Error, don't support global placemenet for tool " + self.eda_tool.value)
        
        
    def run_gp_OpenROAD(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.OPENROAD,
                              step=FlowStep.fixFanout)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.OPENROAD,
                        pre_flow=self.pre_step,
                        step=FlowStep.globalPlace)
        self.build_path()
        
        # run
        run_flow = OpenROADPlacement(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step)
        run_flow.run_global_placement()
        
        return self.run_feature_ieda(reload=True)
    
    #######################################################################################    
    # detail placement
    #######################################################################################
    def run_detail_placement(self):  
        if(self.eda_tool == EdaTool.OPENROAD):
            if os.environ.get('OpenROAD') == "on":
                return self.run_dp_OpenROAD()
            else:
                print("Error, do not support OpenROAD, please install OpenROAD and set OpenROAD environment to 'on'")
                return None
        else:
            print("Error, don't support detail placemenet for tool " + self.eda_tool.value)
        
        
    def run_dp_OpenROAD(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.OPENROAD,
                              step=FlowStep.globalPlace)
        if(self.step == None):
            self.step = DbFlow(eda_tool=EdaTool.OPENROAD,
                        pre_flow=self.pre_step,
                        step=FlowStep.detailPlace)
        self.build_path()
        
        # run openroad
        run_flow = OpenROADPlacement(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog,
                              pre_step = self.pre_step,
                              step = self.step)
        run_flow.run_detail_placement()
        
        return self.run_feature_ieda(reload=True)
    
    #######################################################################################    
    # legalization
    #######################################################################################  
    def run_legalization(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_legalization_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_legalization_innovus()
        
        if(self.eda_tool == EdaTool.DREAMPLACE):
            return self.run_legalization_dreamplace()
        
    def run_legalization_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                              step=FlowStep.optHold)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                        pre_flow=self.pre_step,
                        step=FlowStep.legalization)
        self.build_path()
        
        # run iEDA
        run_flow = IEDAPlacement(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_legalization()
        
        return self.run_feature_ieda()
    
    def run_legalization_innovus(self):
        pass
    
    def run_legalization_dreamplace(self):
        return None
    
    #######################################################################################    
    # filler
    #######################################################################################
    def run_filler(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.run_filler_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.run_filler_innovus()
        
        if(self.eda_tool == EdaTool.DREAMPLACE):
            return self.run_filler_dreamplace()
        
    def run_filler_ieda(self):
        if(self.pre_step == None):
            self.pre_step = DbFlow(eda_tool=EdaTool.IEDA,
                            step=FlowStep.route)
        self.step = DbFlow(eda_tool=EdaTool.IEDA,
                      pre_flow=self.pre_step,
                      step=FlowStep.filler)
        self.build_path()
        
        # run iEDA
        run_flow = IEDAPlacement(dir_workspace = self.dir_workspace,
                              input_def = self.input_def, 
                              output_def = self.output_def, 
                              output_verilog = self.output_verilog)
        run_flow.run_filler()
        
        return self.run_feature_ieda()
    
    def run_filler_innovus(self):
        pass
    
    def run_filler_dreamplace(self):
        pass


#search_space = self._search_space
import os
def main():
    search_space = \
    {'init_wirelength_coef': {'_type': 'uniform', '_value': [0.1, 0.5]}, 'min_wirelength_force_bar': {'_type': 'uniform', '_value': [-500.0, -50.0]}, 'target_density': {'_type': 'uniform', '_value': [0.8, 1.0]}, 'bin_cnt': {'_type': 'choice', '_value': [16, 32, 64, 128, 256, 512, 1024]}, 'max_backtrack': {'_type': 'uniform', '_value': [5, 50]}, 'init_density_penalty': {'_type': 'uniform', '_value': [0.0, 0.001]}, 'target_overflow': {'_type': 'uniform', '_value': [0.0, 0.2]}, 'initial_prev_coordi_update_coef': {'_type': 'uniform', '_value': [50.0, 1000.0]}, 'min_precondition': {'_type': 'uniform', '_value': [1.0, 10.0]}, 'min_phi_coef': {'_type': 'uniform', '_value': [0.75, 1.25]}, 'max_phi_coef': {'_type': 'uniform', '_value': [0.75, 1.25]}}

    from nni.experiment import Experiment
    experiment = Experiment('local')

    experiment.config.trial_command = 'python optimization.py --tuner nni' 
    experiment.config.trial_code_directory = os.path.dirname(__file__)
    experiment.config.search_space = search_space

    experiment.config.tuner.name = "TPE"
    experiment.config.tuner.class_args['optimize_mode'] = "minimize"

    experiment.config.max_trial_number = 2000
    experiment.config.trial_concurrency = 40
    experiment.run(5906)
if __name__=="__main__":
    main()


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File : summary.py
@Author : yell
@Desc : drc database
'''

import sys
import os

from dataclasses import dataclass
from dataclasses import field
from typing import List
from numpy import double, uint, uint64

@dataclass
class StageSummary(object):
    stage: str = None
    die_usage : double = 0.0
    core_usage : double = 0.0
    num_instances : int = 0
    num_nets : int = 0
    num_pins : uint64 = None
    wire_len : double = None

@dataclass
class DesignSummary(object):
    design_name: str = None
    eda_tool : str = None
    die_area : double = 0.0
    core_area : double = 0.0
    num_iopins : int = 0
    num_iopad : int =0
    num_macros : int = 0
    num_layers : int = 0
    num_layers_routing : int = 0
    num_layers_cut : int = 0
    max_fanout : uint = None
    stage_list: List[StageSummary] = field(default_factory=list)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : statis.py
@Author : yell
@Desc : feature statis
'''
from feature.io import FeatureIO
from database.enum import EdaTool, FeatureOption, FlowStep
from workspace.path import WorkspacePath
from flow.flow_db import DbFlow
from database.summary import *
from database.timing import *


class DesignStatis():
    """feature statis"""   
    def __init__(self, dir_workspace : str, eda_tool : EdaTool):
        self.dir_workspace = dir_workspace
        self.eda_tool = eda_tool  
        self.workspace = WorkspacePath(dir_workspace)
        
    def get_statis(self):
        if(self.eda_tool == EdaTool.IEDA):
            return self.get_statis_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.get_statis_innovus()
        
    def get_statis_ieda(self):
        design_statis = self.get_summary_statis()
        
        return design_statis
    
    def get_statis_innovus(self):
        design_statis = self.get_summary_statis()
        
        return design_statis
                     
    def get_summary_statis(self):
        design_summary = DesignSummary()
             
        # get origin feature summary db
        feature_io = FeatureIO(dir_workspace = self.dir_workspace,
                               eda_tool = self.eda_tool, 
                               feature_option = FeatureOption.summary,
                               flow = DbFlow(eda_tool = self.eda_tool,step = FlowStep.place))
        feature_db = feature_io.get_db()
        
        # common statis
        design_summary.design_name = feature_db.info.design_name
        design_summary.eda_tool = self.eda_tool.value
        design_summary.die_area = feature_db.layout.die_area
        design_summary.core_area = feature_db.layout.core_area
        design_summary.num_iopins = feature_db.statis.num_iopins
        design_summary.num_iopad = feature_db.instances.iopads.num
        design_summary.num_macros = feature_db.instances.macros.num
        design_summary.num_layers = feature_db.layers.num_layers
        design_summary.num_layers_routing = feature_db.layers.num_layers_routing
        design_summary.num_layers_cut = feature_db.layers.num_layers_cut
        design_summary.max_fanout = feature_db.pins.max_fanout
        
        # place
        place_summary = StageSummary()
        place_summary.stage = "place"
        place_summary.die_usage = feature_db.layout.die_usage
        place_summary.core_usage = feature_db.layout.core_usage
        place_summary.num_instances = feature_db.statis.num_instances
        place_summary.num_nets = feature_db.nets.num_total
        place_summary.num_pins = feature_db.nets.num_pins
        place_summary.wire_len = feature_db.nets.wire_len
        
        design_summary.stage_list.append(place_summary)
        
        # CTS
        feature_io = FeatureIO(dir_workspace = self.dir_workspace,
                               eda_tool = self.eda_tool, 
                               feature_option = FeatureOption.summary,
                               flow = DbFlow(eda_tool = self.eda_tool,step = FlowStep.cts))
        feature_db = feature_io.get_db()
        
        cts_summary = StageSummary()
        cts_summary.stage = "CTS"
        cts_summary.die_usage = feature_db.layout.die_usage
        cts_summary.core_usage = feature_db.layout.core_usage
        cts_summary.num_instances = feature_db.statis.num_instances
        cts_summary.num_nets = feature_db.nets.num_total
        cts_summary.num_pins = feature_db.nets.num_pins
        cts_summary.wire_len = feature_db.nets.wire_len
        
        design_summary.stage_list.append(cts_summary)
        
        # route
        feature_io = FeatureIO(dir_workspace = self.dir_workspace,
                               eda_tool = self.eda_tool, 
                               feature_option = FeatureOption.summary,
                               flow = DbFlow(eda_tool = self.eda_tool,step = FlowStep.route))
        feature_db = feature_io.get_db()
        
        route_summary = StageSummary()
        route_summary.stage = "route"
        route_summary.die_usage = feature_db.layout.die_usage
        route_summary.core_usage = feature_db.layout.core_usage
        route_summary.num_instances = feature_db.statis.num_instances
        route_summary.num_nets = feature_db.nets.num_total
        route_summary.num_pins = feature_db.nets.num_pins
        route_summary.wire_len = feature_db.nets.wire_len
        
        design_summary.stage_list.append(route_summary)
        
        return design_summary        


from setuptools import setup, find_packages
import os
from pathlib import Path

# 读取版本信息
def get_version():
    init_file = Path(__file__).parent / "aieda" / "__init__.py"
    if init_file.exists():
        with open(init_file) as f:
            for line in f:
                if line.startswith("__version__"):
                    return line.split("=")[1].strip().strip('"\'')
    return "0.1.0"

# 读取README
def get_long_description():
    readme_file = Path(__file__).parent / "README.md"
    if readme_file.exists():
        return readme_file.read_text(encoding="utf-8")
    return "AI-Enhanced Electronic Design Automation Library with iEDA Integration"

# 读取requirements
def get_requirements():
    requirements_file = Path(__file__).parent / "requirements.txt"
    if requirements_file.exists():
        with open(requirements_file) as f:
            return [line.strip() for line in f if line.strip() and not line.startswith("#")]
    
    # 默认依赖
    return [
        "numpy>=1.20.0",
        "pandas>=1.3.0",
        "pyyaml>=5.4.0",
        "click>=8.0.0",
        "tqdm>=4.60.0",
    ]

setup(
    name="aieda",
    version=get_version(),
    author="yhqiu",
    author_email="qiuyihang23@mails.ucas.ac.cn",
    description="AI-Enhanced Electronic Design Automation Library with iEDA Integration",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/OSCC-Project/AiEDA",
    
    packages=find_packages(),
    
    # 包含所有数据文件
    package_data={
    },
    include_package_data=True,
    
    python_requires=">=3.8",
    install_requires=get_requirements(),
    
    # 可选依赖
    extras_require={
        # "dev": [
        #     "pytest>=6.0",
        #     "black>=21.0",
        #     "flake8>=3.8",
        #     "mypy>=0.800",
        # ],
        # "full": [
        #     "matplotlib>=3.3.0",
        #     "plotly>=5.0.0",
        #     "jupyter>=1.0.0",
        # ]
    },
    
    # 命令行工具
    entry_points={
        "console_scripts": [
            "aieda-flow=aieda.application.flow.run_iEDA_flow:main"
            # "aieda=aieda.cli:main",  # 如果你有CLI模块的话
        ],
    },
    
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research", 
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    
    keywords="eda, ieda, vlsi, synthesis, place-and-route, ai",
)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File : lm_graph.py
@Author : yell
@Desc : feature for large model
"""

from database.enum import EdaTool, FeatureOption, FlowStep
from workspace.path import WorkspacePath, LargeModelFeatureType
from flow.flow_db import DbFlow
from database.large_model import *
from utility.json_parser import JsonParser
from typing import List, Dict, Tuple, Any
from math import gcd
from collections import defaultdict
import pandas as pd
from tqdm import tqdm
import networkx as nx
import json
from dataclasses import asdict
import yaml
import orjson
import os
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, as_completed
import os


class LmGraph:
    """feature statis"""

    def __init__(self, dir_workspace: str, eda_tool: EdaTool):
        self.dir_workspace = dir_workspace
        self.eda_tool = eda_tool
        self.workspace = WorkspacePath(dir_workspace)

    def get_graph(self) -> List[LmNet]:
        flow = DbFlow(eda_tool=self.eda_tool, step=FlowStep.route)
        files = list(self._iter_json_files(self.workspace.get_feature_lm(
            flow=flow,
            feature_option=FeatureOption.large_model,
            lm_feature_type=LargeModelFeatureType.lm_nets,
        )))
        lm_nets = []
        for fp in tqdm(files, desc="Parsing nets batch file"):
            lm_nets.extend(self._parse_nets(fp))
        return lm_nets

        # ctx = mp.get_context('fork')
        # with ProcessPoolExecutor(max_workers=os.cpu_count(),
        #                          mp_context=ctx) as exe:
        #     lm_nets = list(tqdm(exe.map(self._parse_nets, files), total=len(files),
        #                         desc="Parsing nets batch file"))
        # return lm_nets

    @staticmethod
    def _iter_json_files(root: str):
        for ent in os.scandir(root):
            if ent.is_dir():
                yield from LmGraph._iter_json_files(ent.path)
            elif ent.name.endswith(".json"):
                yield ent.path

    def _parse_nets(self, batch_filepath: str) -> List[LmNet]:
        with open(batch_filepath, "rb") as f:
            lm_nets_metadata = orjson.loads(f.read())
        lm_nets = []
        for net_metadata in lm_nets_metadata:
            lm_net = self._parse_net(net_metadata)
            if lm_net:
                lm_nets.append(lm_net)
        return lm_nets

    def _parse_net(self, net_metadata: Dict[str, Any]) -> LmNet:
        lm_net = LmNet()
        lm_net.id = net_metadata.get("id")
        lm_net.name = net_metadata.get("name")

        # net feature
        net_feature = LmNetFeature()
        feature_data = net_metadata.get('feature', {})
        net_feature.llx = feature_data.get('llx')
        net_feature.lly = feature_data.get('lly')
        net_feature.urx = feature_data.get('urx')
        net_feature.ury = feature_data.get('ury')
        net_feature.wire_len = feature_data.get('wire_len')
        net_feature.via_num = feature_data.get('via_num')
        net_feature.drc_num = feature_data.get('drc_num')
        net_feature.R = feature_data.get('R')
        net_feature.C = feature_data.get('C')
        net_feature.power = feature_data.get('power')
        net_feature.delay = feature_data.get('delay')
        net_feature.slew = feature_data.get('slew')
        net_feature.fanout = feature_data.get('fanout')

        if 'aspect_ratio' in feature_data:
            net_feature.aspect_ratio = feature_data.get('aspect_ratio')
            net_feature.width = feature_data.get('width')
            net_feature.height = feature_data.get('height')
            net_feature.area = feature_data.get('area')
            net_feature.l_ness = feature_data.get('l_ness')
            net_feature.drc_type = feature_data.get('drc_type')
            net_feature.volume = feature_data.get('volume')
            net_feature.layer_ratio = feature_data.get('layer_ratio')
            net_feature.rsmt = feature_data.get('rsmt')

        lm_net.feature = net_feature

        # pins
        lm_net.pin_num = net_metadata.get("pin_num", 0)
        json_pins = net_metadata.get("pins", [])
        for json_pin in json_pins:
            lm_pin = LmPin()
            lm_pin.id = json_pin.get("id")
            lm_pin.instance = json_pin.get("i")
            lm_pin.pin_name = json_pin.get("p")
            lm_pin.is_driver = json_pin.get("driver")

            lm_net.pins.append(lm_pin)

        # wires
        lm_net.wire_num = net_metadata.get("wire_num", 0)
        json_wires = net_metadata.get("wires", [])
        for json_wire in json_wires:
            lm_wire = LmWire()
            lm_wire.id = json_wire.get("id")

            # wire feature
            wire_feature = LmWireFeature()
            wire_feature_data = json_wire.get("feature", {})
            wire_feature.wire_width = wire_feature_data.get("wire_width")
            wire_feature.wire_len = wire_feature_data.get("wire_len")
            wire_feature.drc_num = wire_feature_data.get("drc_num")
            wire_feature.R = wire_feature_data.get("R")
            wire_feature.C = wire_feature_data.get("C")
            wire_feature.power = wire_feature_data.get("power")
            wire_feature.delay = wire_feature_data.get("delay")
            wire_feature.slew = wire_feature_data.get("slew")
            wire_feature.congestion = wire_feature_data.get("congestion")
            wire_feature.wire_density = wire_feature_data.get(
                "wire_density")
            wire_feature.drc_type = wire_feature_data.get("drc_type")

            lm_wire.feature = wire_feature

            # wire connections
            wire_data = json_wire.get("wire", {})
            wire_connections = LmPath()

            lm_node1 = LmNode()
            lm_node1.id = wire_data.get("id1")
            lm_node1.x = wire_data.get("x1")
            lm_node1.y = wire_data.get("y1")
            lm_node1.real_x = wire_data.get("real_x1")
            lm_node1.real_y = wire_data.get("real_y1")
            lm_node1.row = wire_data.get("r1")
            lm_node1.col = wire_data.get("c1")
            lm_node1.layer = wire_data.get("l1")
            lm_node1.pin_id = wire_data.get("p1")
            wire_connections.node1 = lm_node1

            lm_node2 = LmNode()
            lm_node2.id = wire_data.get("id2")
            lm_node2.x = wire_data.get("x2")
            lm_node2.y = wire_data.get("y2")
            lm_node2.real_x = wire_data.get("real_x2")
            lm_node2.real_y = wire_data.get("real_y2")
            lm_node2.row = wire_data.get("r2")
            lm_node2.col = wire_data.get("c2")
            lm_node2.layer = wire_data.get("l2")
            lm_node2.pin_id = wire_data.get("p2")
            wire_connections.node2 = lm_node2

            lm_wire.wire = wire_connections

            # path
            lm_wire.path_num = json_wire.get("path_num", 0)
            json_paths = json_wire.get("paths", [])
            for json_path in json_paths:
                wire_path = LmPath()

                lm_path_node1 = LmNode()
                lm_path_node1.id = json_path.get("id1")
                lm_path_node1.x = json_path.get("x1")
                lm_path_node1.y = json_path.get("y1")
                lm_path_node1.real_x = json_path.get("real_x1")
                lm_path_node1.real_y = json_path.get("real_y1")
                lm_path_node1.row = json_path.get("r1")
                lm_path_node1.col = json_path.get("c1")
                lm_path_node1.layer = json_path.get("l1")
                wire_path.node1 = lm_path_node1

                lm_path_node2 = LmNode()
                lm_path_node2.id = json_path.get("id2")
                lm_path_node2.x = json_path.get("x2")
                lm_path_node2.y = json_path.get("y2")
                lm_path_node2.real_x = json_path.get("real_x2")
                lm_path_node2.real_y = json_path.get("real_y2")
                lm_path_node2.row = json_path.get("r2")
                lm_path_node2.col = json_path.get("c2")
                lm_path_node2.layer = json_path.get("l2")
                wire_path.node2 = lm_path_node2

                lm_wire.paths.append(wire_path)

            lm_net.wires.append(lm_wire)

        # routing graph
        routing_graph_data = net_metadata.get("routing_graph", {})
        vertices = []
        for v in routing_graph_data.get("vertices", []):
            point = LmNetRoutingPoint(
                x=v["x"],
                y=v["y"],
                layer_id=v["layer_id"]
            )
            vertex = LmNetRoutingVertex(
                id=v["id"],
                is_pin=v["is_pin"],
                is_driver_pin=v["is_driver_pin"],
                point=point
            )
            vertices.append(vertex)

        edges = []
        for e in routing_graph_data.get("edges", []):
            path = [LmNetRoutingPoint(**p) for p in e["path"]]
            edge = LmNetRoutingEdge(source_id=e["source_id"],
                                    target_id=e["target_id"], path=path)
            edges.append(edge)
        routing_graph = LmNetRoutingGraph(vertices=vertices, edges=edges)
        lm_net.routing_graph = routing_graph

        return lm_net


class LmWirePatternGen:
    def __init__(self, epsilon: int = 1):
        self._epsilon = epsilon
        self._patterns: Dict[str, LmWirePatternSeq] = {}
        self._pattern_count: Dict[str, int] = defaultdict(int)

    def pattern_summary(self, csv_path: str = None) -> pd.DataFrame:
        df = pd.DataFrame(self._pattern_count.items(),
                          columns=["Pattern", "Count"])
        df.sort_values(by="Count", ascending=False,
                       inplace=True, ignore_index=True)
        if csv_path:
            df.to_csv(csv_path, index=False)
        return df

    def add_wire(self, wire: LmWire) -> LmWirePatternSeq:
        pattern = self.gen_pattern(wire)
        self.add_pattern(pattern)
        return pattern

    def add_pattern(self, pattern: LmWirePatternSeq):
        if pattern.name in self._pattern_count:
            self._pattern_count[pattern.name] += 1
        else:
            self._patterns[pattern.name] = pattern
            self._pattern_count[pattern.name] = 1

    def gen_pattern(self, wire: LmWire) -> LmWirePatternSeq:
        point_list = self._get_point_list(wire)
        pattern = self._calc_pattern(point_list)
        return pattern

    def _get_point_list(self, wire: LmWire) -> List[LmWirePatternPoint]:
        paths = wire.paths
        points = [
            LmWirePatternPoint(path.node1.x, path.node1.y, path.node1.layer)
            for path in paths
        ]
        end = LmWirePatternPoint(
            paths[-1].node2.x, paths[-1].node2.y, paths[-1].node2.layer
        )
        points.append(end)
        return points

    def _calc_pattern(self, points: List[LmWirePatternPoint]) -> LmWirePatternSeq:
        sorted_points = points[:]
        if sorted_points[0].x > sorted_points[-1].x or (
            sorted_points[0].x == sorted_points[-1].x
            and sorted_points[0].y > sorted_points[-1].y
        ):
            sorted_points.reverse()

        pattern = LmWirePatternSeq()
        for i in range(len(sorted_points) - 1):
            start = sorted_points[i]
            end = sorted_points[i + 1]
            x_same = start.x == end.x
            y_same = start.y == end.y
            z_same = start.z == end.z
            if x_same and y_same and z_same:
                continue

            if x_same and y_same:
                direction = LmWirePatternDirection.VIA
                length = 1
            elif x_same:
                direction = (
                    LmWirePatternDirection.TOP
                    if start.y < end.y
                    else LmWirePatternDirection.BOTTOM
                )
                length = abs(start.y - end.y)
            else:
                direction = (
                    LmWirePatternDirection.RIGHT
                    if start.x < end.x
                    else LmWirePatternDirection.LEFT
                )
                length = abs(start.x - end.x)

            pattern.units.append(LmWirePatternUnit(direction, length))

        if not pattern.units:
            return pattern

        max_common_factor = pattern.units[0].length
        for unit in pattern.units:
            max_common_factor = gcd(max_common_factor, unit.length)

        for unit in pattern.units:
            unit.length //= max_common_factor

        pattern_name = ""
        for unit in pattern.units:
            direction = unit.direction.name[0]
            normalized_length = unit.length // self._epsilon + 1
            pattern_name += f"{direction}{normalized_length}"

        pattern.name = pattern_name
        return pattern


class LmNetsPatternsGen:
    def __init__(self, epsilon: int = 1):
        self._gen = LmWirePatternGen(epsilon=epsilon)

    def gen(self, lm_nets: List[LmNet]):
        for lm_net in tqdm(lm_nets):
            wires = lm_net.wires
            for wire in wires:
                self._gen.add_wire(wire)

    def summary(self, csv_path: str = None):
        return self._gen.pattern_summary(csv_path)


class LmNetSeqConverter:
    def __init__(self, epsilon: int = 1):
        self._gen = LmWirePatternGen(epsilon=epsilon)
        self._seqs = []

    def build_seqs(self, lm_nets: List[LmNet]) -> List[LmNetSeq]:
        for lm_net in lm_nets:
            graph = self._convert(lm_net)
            seqs = self._convert_to_seq(graph)
            self._seqs.extend(seqs)
        return self._seqs

    def save_seqs(self, json_path: str):
        with open(json_path, "w") as f:
            json.dump(self._seqs, f, default=asdict)

    def load_seqs(self, json_path: str) -> List[LmNetSeq]:
        with open(json_path, "r") as f:
            self._seqs = json.load(f)
        return self._seqs

    def _convert(self, lm_net: LmNet) -> nx.Graph:
        # edge with pattern
        graph = nx.Graph()
        wires = lm_net.wires
        for wire in wires:
            wire: LmWire
            start = wire.wire.node1
            end = wire.wire.node2

            pattern = self._gen.add_wire(wire)
            pattern_name = pattern.name
            if start.id not in graph:
                graph.add_node(start.id, pos=(start.x, start.y, start.layer))
            if end.id not in graph:
                graph.add_node(end.id, pos=(end.x, end.y, end.layer))
            graph.add_edge(start.id, end.id, pattern=pattern_name)

        return graph

    def _convert_to_seq(self, net_graph: nx.Graph) -> List[LmNetSeq]:
        seqs = []
        source = 0
        for node in net_graph.nodes:
            if net_graph.degree(node) == 1:
                source = node
                break
        targets = []
        for node in net_graph.nodes:
            if net_graph.degree(node) == 1 and node != source:
                targets.append(node)
        for target in targets:
            loc_seq = []
            pattern_seq = []
            path = nx.shortest_path(net_graph, source, target)
            for vertex in path:
                pos = net_graph.nodes[vertex]["pos"]
                loc_seq.append(LmWirePatternPoint(pos[0], pos[1], pos[2]))
            for i in range(1, len(path)):
                pattern = net_graph[path[i - 1]][path[i]]["pattern"]
                pattern_seq.append(pattern)
            seq = LmNetSeq(loc_seq, pattern_seq)
            seqs.append(seq)
        return seqs


class LmTimingGraphSeqConverter:
    def __init__(self, epsilon: int = 1):
        self._gen = LmWirePatternGen(epsilon=epsilon)
        self._seqs = []

    def convert(self, timing_wire_graph, lm_nets: List[LmNet]) -> nx.Graph:
        nodes = timing_wire_graph.nodes
        edges = timing_wire_graph.edges
        graph = nx.Graph()
        # build nodes
        for node in nodes:
            graph.add_node(node.name, is_pin=node.is_pin, is_port=node.is_port)
        # build edges
        wire_map = self._build_wire_map(lm_nets)
        for edge in edges:
            from_node = nodes[edge.from_node]
            to_node = nodes[edge.to_node]
            key = (
                (int)(from_node.name.split(":")[1]),
                (int)(to_node.name.split(":")[1]),
            )
            pattern = "" if key not in wire_map else self._gen.add_wire(
                wire_map[key])
            graph.add_edge(
                from_node.name,
                to_node.name,
                feature_R=edge.feature_R,
                feature_C=edge.feature_C,
                feature_from_slew=edge.feature_from_slew,
                feature_to_slew=edge.feature_to_slew,
                is_net_edge=edge.is_net_edge,
                pattern=pattern,
            )
        return graph

    def convert_to_seq(self, graph: nx.Graph) -> List[LmNetSeq]:
        seqs = []
        pass

    def _build_wire_map(self, lm_nets: List[LmNet]) -> Dict[Tuple[int, int], LmWire]:
        wire_map = {}
        for lm_net in lm_nets:
            wires = lm_net.wires
            for wire in wires:
                wire: LmWire
                start = wire.wire.node1.id
                end = wire.wire.node2.id
                wire_map[(start, end)] = wire
                wire_map[(end, start)] = wire
        return wire_map


class LmTimingWireGraphBuilder:
    """The parser for LM wire timing graph."""

    def __init__(self, dir_workspace: str, eda_tool: EdaTool):
        self.dir_workspace = dir_workspace
        self.eda_tool = eda_tool
        self.workspace = WorkspacePath(dir_workspace)

    def get_wire_graph(self) -> LmTimingWireGraph:

        flow = DbFlow(eda_tool=self.eda_tool, step=FlowStep.route)
        feature_dir = self.workspace.get_feature_lm(
            flow=flow,
            feature_option=FeatureOption.large_model,
            lm_feature_type=LargeModelFeatureType.lm_wire_graph,
        )

        yaml_path = feature_dir + "timing_wire_graph.yaml"

        if not os.path.exists(yaml_path):
            return None

        with open(yaml_path, "r") as file:
            print("load wire graph yaml start")
            wire_nodes = []
            wire_edges = []

            # not use yaml module
            # wire_graph_data = yaml.safe_load(file)
            # print("load wire graph yaml end")
            # # wire graph
            # for item_name, item in wire_graph_data.items():
            #     if "node" in item_name:
            #         wire_node = LmTimingWireGraphNode(item["name"], item["is_pin"], item["is_port"])
            #         wire_nodes.append(wire_node)

            #     if "edge" in item_name:
            #         from_index = item["from_node"]
            #         to_index = item["to_node"]
            #         wire_edge = LmTimingWireGraphEdge(from_index, to_index)
            #         wire_edges.append(wire_edge)

            wire_node = None
            wire_edge = None
            edge_set = set()
            for _, line in tqdm(
                enumerate(file), desc="Processing load wire graph yaml"
            ):
                line = line.strip()
                if line.startswith("node_"):
                    if wire_node is not None:
                        wire_nodes.append(wire_node)
                        wire_node = None
                elif line.startswith("edge_"):
                    # add last node
                    if wire_node is not None:
                        wire_nodes.append(wire_node)
                        wire_node = None

                    if wire_edge is not None:
                        if not (wire_edge.from_node, wire_edge.to_node) in edge_set:
                            wire_edges.append(wire_edge)
                            edge_set.add(
                                (wire_edge.from_node, wire_edge.to_node))
                        wire_edge = None
                else:
                    # Split at the first colon only
                    key, value = line.split(":", 1)
                    if key == "name":
                        wire_node = LmTimingWireGraphNode(
                            value.strip(), False, False)
                    elif key == "is_pin":
                        wire_node.is_pin = True if int(
                            value.strip()) == 1 else False
                    elif key == "is_port":
                        wire_node.is_port = True if int(
                            value.strip()) == 1 else False
                    elif key == "from_node":
                        wire_edge = LmTimingWireGraphEdge(
                            int(value.strip()), None)
                    elif key == "to_node":
                        wire_edge.to_node = int(value.strip())

            # add last edge
            wire_edges.append(wire_edge)
            wire_timing_graph = LmTimingWireGraph(wire_nodes, wire_edges)

            print("load wire graph yaml end")
            print("wire graph nodes num: ", len(wire_nodes))
            print("wire graph edges num: ", len(wire_edges))
            return wire_timing_graph

        return None



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : run_pr_flow.py
@Author : yell
@Desc : run all flow 
'''
import os
from flow.flow_manager import FlowManager
from flow.flow_db import DbFlow
from database.enum import EdaTool, FlowStep
from tools.iEDA.feature.feature import IEDAFeature
from workspace.path import WorkspacePath
from multiprocessing import Process
from engine.floorplan import EngineFloorplan
from engine.cts import EngineCTS
from engine.placement import EnginePlacement
from engine.routing import EngineRouting
from engine.net_opt import EngineNetOpt # input_def = "/data/project_share/dataset/t28/gcd/gcd_floorplan.def"
from engine.timing_opt import EngineTimingOpt
from engine.gds import EngineGDS
from engine.drc import EngineDRC

class RunFlowIEDA:
    '''run eda backend flow
    '''
    
    def __init__(self, dir_workspace : str):        
        self.dir_workspace = dir_workspace
        
        self.workspace = WorkspacePath(dir_workspace)
        self.flow_manager = FlowManager(self.workspace.get_config('flow'))
    
    def run(self):
        """run flow"""
        flow = self.flow_manager.get_new_flow()
        while flow is not None :
            #run flow 
            self.flow_manager.set_current_flow(flow)
            if self.run_flow(flow) is False:
                return False
            #set as pre flow
            self.flow_manager.set_pre_flow(flow)
            #get next flow
            flow = self.flow_manager.get_new_flow()
          
        #check all flow success  
        for check_flow in self.flow_manager.flow_list :
            if check_flow.is_finish() is False:
                return False
        
        return True
    
    def reset_flow_state(self):
        self.flow_manager.reset_flow_state()

            
    def run_flow(self, flow : DbFlow):
        """run flow"""    
        #set state running
        flow.set_state_running()
        self.flow_manager.save_flow_state(flow)
             
        #run eda tool
        self.run_eda_process(flow)
        
        #save flow state
        if self.check_flow_state(flow) is True:
            flow.set_state_finished()
            self.flow_manager.save_flow_state(flow)
            return True
        else:
            flow.set_state_imcomplete()   
            self.flow_manager.save_flow_state(flow)
            return False 
    
    def run_eda_process(self, flow : DbFlow):   
        """run eda as process"""
        p = Process(target=self.run_eda_tool, args=(self.workspace, flow,))
        p.start()
        p.join()
            
    def run_eda_tool(self, workspace : WorkspacePath, flow : DbFlow):
        io_files = self.build_io_file(flow)
        
        """run eda tool"""  
        if (flow.eda_tool == EdaTool.IEDA):
            # run iEDA
            self.build_flow(flow.step, io_files)
            
            print(io_files)
            
            # # save feature
            # ieda_feature = IEDAFeature(workspace.workspace)
            # ieda_feature.feature_summary(flow, True)
        else:
            pass
    
    def build_io_file(self, flow : DbFlow):
        io_files = {}
        
        io_files["input_def"] = self.workspace.get_input_def(flow)
        io_files["input_verilog"] = self.workspace.get_input_verilog(flow)
            
        io_files["output_def"] = self.workspace.get_output_def(flow=flow, compressed=True)
        io_files["output_verilog"] = self.workspace.get_output_verilog(flow=flow, compressed=True)
        
        return io_files
       
    def build_flow(self, step: FlowStep, io_files : dict):        
        if(step == FlowStep.floorplan):
            ieda_flow = EngineFloorplan(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run()
        
        if(step == FlowStep.fixFanout):
            ieda_flow = EngineNetOpt(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run()
            
        if(step == FlowStep.place):
            ieda_flow = EnginePlacement(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run_placer()
        
        if(step == FlowStep.cts):
            ieda_flow = EngineCTS(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run()
            
        if(step == FlowStep.optDrv):
            ieda_flow = EngineTimingOpt(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run_to_drv()
            
        if(step == FlowStep.optHold):
            ieda_flow = EngineTimingOpt(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run_to_hold()
            
        if(step == FlowStep.optSetup):
            ieda_flow = EngineTimingOpt(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run_to_setup()
            
        if(step == FlowStep.legalization):
            ieda_flow = EnginePlacement(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run_legalization()
            
        if(step == FlowStep.route):
            ieda_flow = EngineRouting(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run_route()
            
        if(step == FlowStep.filler):
            ieda_flow = EnginePlacement(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"],
                                      output_def = io_files["output_def"],
                                      output_verilog = io_files["output_verilog"])
            ieda_flow.run_filler()
            
        if(step == FlowStep.gds):
            ieda_flow = EngineGDS(self.dir_workspace,input_def = io_files["input_def"])
            ieda_flow.run()
        
        if(step == FlowStep.drc):
            ieda_flow = EngineDRC(dir_workspace = self.dir_workspace,
                                      input_def = io_files["input_def"],
                                      input_verilog = io_files["input_verilog"])
            ieda_flow.run()
        
    
    def check_flow_state(self, flow : DbFlow):
        """check state"""
        #check flow success if output def & verilog file exist
        output_def = self.workspace.get_output_def(flow=flow, compressed=True)
        output_verilog = self.workspace.get_output_verilog(flow=flow, compressed=True)
        
        return (os.path.exists(output_def) and os.path.exists(output_verilog))


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : lm_path.py
@Time : 2025/01/20 15:04:36
@Author : simin tao
@Version : 1.0
@Contact : taosm@pcl.ac.cn, liuh0326@163.com
@Desc : 
'''
import yaml
import torch
import hashlib

from database.enum import EdaTool, FeatureOption, FlowStep
from workspace.path import WorkspacePath, LargeModelFeatureType
from database.large_model import LmTimingWireGraphNode, LmTimingWireGraphEdge, LmTimingWirePathGraph

class LmTimingWirePathGraphBuilder:
    def __init__(self, file_path, eda_tool: EdaTool = None, step = None):
        # for timing predict
        self.file_path = file_path
        self.data = None
        self.capacitance_list = []
        self.slew_list = []
        self.resistance_list = []
        self.incr_list = []
        
        # record path nodes
        self.nodes = []
        
        # for flow control
        self.eda_tool = eda_tool
        self.step = step

    def load_data(self):
        """Load YAML data from the file."""
        try:
            with open(self.file_path, 'r') as f:
                self.data = yaml.safe_load(f)
        except Exception as e:
            print(f"Error loading file: {e}")

    def parse_data(self):
        """Parse nodes, net arcs, and instance arcs in order."""
        remove_parentheses_content = lambda s: s[:s.find('(')].strip() if s.find('(')!= -1 else s
        for key, value in self.data.items():
            if key.startswith("node_"):
                # Parse node
                self.capacitance_list.append(value.get("Capacitance", 0))
                self.slew_list.append(value.get("slew", 0))
                self.resistance_list.append(0)  # Default R value for nodes
                
                # record pin node
                node_name = value.get("Point", "")
                # remove instance cell name
                node_name = remove_parentheses_content(node_name)
                if not len(self.nodes) or node_name != self.nodes[-1]: 
                    self.nodes.append(node_name)
                    
            elif key.startswith("net_arc_"):
                self.incr_list.append(value.get("Incr", 0))
                for edge_key, edge_value in value.items():
                    if edge_key.startswith("edge_"):
                        self.capacitance_list.append(edge_value.get("wire_C", 0))
                        self.slew_list.append(edge_value.get("to_slew", 0))
                        self.resistance_list.append(edge_value.get("wire_R", 0))
                        
                        # record edge node
                        self.nodes.append(edge_value.get("wire_to_node", ""))
  
            elif key.startswith("inst_arc_"):
                self.incr_list.append(value.get("Incr", 0))


    def get_combined_tensor(self):
        """Combine all lists into a single 2D tensor with each list as a row."""
        combined_data = [
            self.capacitance_list,
            self.slew_list,
            self.resistance_list
        ]
        tensor = torch.tensor(combined_data, dtype=torch.float32)
        return tensor

    def get_incr_tensor(self):
        """Get the tensor of Incr values and calculate the sum."""
        incr_tensor = torch.tensor(self.incr_list, dtype=torch.float32)
        incr_sum = incr_tensor.sum().item()
        return incr_tensor, incr_sum

    @staticmethod
    def pad_tensors(tensor_list, max_length):
        """Pad all tensors in the list to the max length."""
        padded_tensors = []
        for tensor in tensor_list:
            padded = torch.nn.functional.pad(tensor, (0, max_length - tensor.size(1)), "constant", 0)
            padded_tensors.append(padded)
        return padded_tensors
    
    def generate_hash(self):
        """Generate a hash for the concatenated unique strings."""
        concatenated = "".join(self.nodes)
        hash_object = hashlib.md5(concatenated.encode())
        return hash_object.hexdigest() 

    
    def construct_path_graph(self) -> LmTimingWirePathGraph:
        r"""construct a path graph from yaml data.
        """
        wire_path_nodes = []
        wire_path_edges = []
        for index, node_name in enumerate(self.nodes):
            parts = node_name.split(":")
            is_port = True if len(parts) == 1 else False
            is_pin = True if len(parts) == 2 and not parts[1].isdigit() else False
            wire_path_node = LmTimingWireGraphNode(node_name, is_pin, is_port)
            wire_path_nodes.append(wire_path_node)
            
            if index > 0:
                wire_path_edge = LmTimingWireGraphEdge(index - 1, index)
                wire_path_edges.append(wire_path_edge)
                
        wire_path_graph = LmTimingWirePathGraph(wire_path_nodes, wire_path_edges)
        return wire_path_graph



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
'''
@File : main.py
@Time : 2022/07/11 10:23:08
@Author : shuaiying long
@Version : 1.0
@Contact : longshy@pcl.ac.cn
@Desc : The top script of ibm project.
'''


# from writer.writer import Writer
# from parser.parser import Parser, TimingPathParser, FanloadViolatParser
# from tkinter import Tcl
# import sys
# import os
# BASE = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
# sys.path.insert(0, BASE)


# def run_netparser():
#     filename_reader = "/home/longshuaiying/iBM/tests/example/report_command_rpt/result/reportWire.rpt"
#     filename_writer = "/home/longshuaiying/iBM/ibm/report/net_wire_length.csv"
#     parser = Parser()
#     # reader = Reader()
#     writer = Writer()
#     net_list = parser.net_parser(filename_reader)
#     writer.net_writer(filename_writer, net_list)


# def run_hinstparser():
#     filename_reader = "/home/longshuaiying/iBM/tests/example/report_command_rpt/result/report_area_detail.rpt"
#     filename_writer = "/home/longshuaiying/iBM/ibm/report/area.csv"
#     parser = Parser()
#     writer = Writer()
#     hinst_list = parser.hinst_parser(filename_reader)
#     writer.hinst_writer(filename_writer, hinst_list)


# def run_timingparser():
#     filename_reader = "/home/longshuaiying/run_data/work_0/beihai_Flow/pd_data/pr/rpt/i-perf/iCTS/2022-08-02/innovus/CTS/timingReports/asic_top_postCTS_all_hold.tarpt"
#     filename_writer = "/home/longshuaiying/iBM/ibm/report/hold_timingpath.csv"
#     timingpath_parser = TimingPathParser(filename_reader)
#     concate_timing_path_list = []
#     concate_timing_path_list = timingpath_parser.parser_concatetimingpath()
#     writer = Writer()
#     writer.timingpath_writer(filename_writer, concate_timing_path_list)


# def run_fanloadparser():
#     filename_reader = "/home/longshuaiying/run_data/work_0/beihai_Flow/pd_data/pr/rpt/i-perf/iCTS/2022-08-02/innovus/CTS/timingReports/asic_top_postCTS.fanout"
#     filename_writer = "/home/longshuaiying/iBM/ibm/report/fanload_violat.csv"
#     fanloadviolat_parser = FanloadViolatParser(filename_reader)
#     fanout_load_violat_info = fanloadviolat_parser.parser_fanoutloadviolat()
#     writer = Writer()
#     writer.fanloadviolat_writer(filename_writer, fanout_load_violat_info)


# def run_sourcetcl():
#     tcl = Tcl()
#     run_paths: list[str] = ["source /home/longshuaiying/iBM/ibm/fpt/command/Stage1.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage2.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage3.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage4.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage5.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage6.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage7.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage8.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage9.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage10.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage11.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage12.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage13.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage14.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage15.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage16.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage17.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage18.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage19.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage20.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage21.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage22.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage23.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage24.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage25.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage26.tcl",
#                             "source /home/longshuaiying/iBM/ibm/fpt/command/Stage27.tcl"]
#     for run_path in run_paths:
#         tcl.eval(run_path)


if __name__ == '__main__':
    # run_timingparser()
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
######################################################################################
# import ai-eda as root

######################################################################################
import sys
import os
import argparse
from typing import Optional

# set EDA tools working environment
# option : iEDA innovus PT dreamPlace
os.environ['eda_tool'] = "iEDA"

current_dir = os.path.split(os.path.abspath(__file__))[0]
aieda_dir = current_dir.rsplit('/', 2)[0]
sys.path.append(aieda_dir)

print(f"current_dir = {current_dir}")
print(f"aieda_dir = {aieda_dir}")

from feature.io import FeatureIO
from database.enum import EdaTool, FlowStep, FeatureOption
from flow.flow_db import DbFlow


def run_feature_evaluation(design: str, workspace_path: str, step: Optional[str] = None) -> None:
    """
    运行特征评估
    
    参数:
        design: 设计名称
        workspace_path: 工作空间路径，包含{design}占位符
        step: 要评估的流程步骤, 默认为place
    
    返回:
        None
    """
    # 构建工作空间路径
    dir_workspace = workspace_path.format(design=design)
    if not os.path.exists(dir_workspace):
        print(f"错误: 工作空间不存在 - {dir_workspace}")
        return
    
    # 确定要使用的流程步骤
    flow_step = FlowStep.place  # 默认为place
    
    # 如果指定了步骤，则尝试转换为FlowStep枚举值
    if step:
        try:
            # 将字符串转换为FlowStep枚举
            flow_step = getattr(FlowStep, step) if step in dir(FlowStep) else FlowStep[step]
            print(f"使用指定步骤: {flow_step.value}")
        except (AttributeError, KeyError):
            # 提示有效的选项并退出程序
            valid_steps = [attr for attr in dir(FlowStep) 
                          if not attr.startswith('_') 
                          and attr != 'value' 
                          and attr != 'name']
            print(f"错误: 无效的步骤 '{step}'")
            print(f"有效的步骤包括: {', '.join(valid_steps)}")
            sys.exit(1)  # 使用非零退出码表示错误
    else:
        print(f"使用默认步骤: {flow_step.value}")
    
    try:
        # 创建FeatureIO对象并生成评估
        metrics_eval = FeatureIO(
            dir_workspace=dir_workspace,
            eda_tool=EdaTool.IEDA, 
            feature_option=FeatureOption.eval,
            flow=DbFlow(eda_tool=EdaTool.IEDA, step=flow_step)
        )
        metrics_eval.generate(reload=True)
        print(f"特征评估完成！")
    except Exception as e:
        print(f"错误: 生成特征评估时出错 - {str(e)}")

if __name__ == "__main__":
    # 创建参数解析器
    parser = argparse.ArgumentParser(description='指标评估工具')
    parser.add_argument('--design', type=str, default='gcd', 
                        help='设计名称, 默认值为gcd')
    parser.add_argument('--path', type=str, default='/data/project_share/yhqiu/{design}/workspace', 
                        help='工作空间路径，需包含{design}占位符，默认值为/data/project_share/yhqiu/{design}/workspace')
    parser.add_argument('--step', type=str, default=None, 
                        help='要评估的流程步骤, 默认为place。目前可选: place, cts')
    args = parser.parse_args()
    
    # 特征评估
    run_feature_evaluation(design=args.design, workspace_path=args.path, step=args.step)
    
    exit(0)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : flow_manager.py
@Author : yell
@Desc : Manage flow
'''
from dataclasses import dataclass

from database.enum import FlowStep, EdaTool, PTState, TaskState, FptState

@dataclass
class DbFlow(object):
    """flow data structure"""
    eda_tool : EdaTool  = None
    pre_flow : any = None
    step : FlowStep = FlowStep.NoStep
    state : TaskState = TaskState.Unstart
    fpt_state : FptState = FptState.on
    pt_state : PTState = PTState.off
    is_first : bool = False

    def set_state_unstart(self):
        """set_state_unstart"""
        self.state = TaskState.Unstart
    
    def set_state_running(self):
        """set_state_running"""
        self.state = TaskState.Ongoing
        
    def set_state_finished(self):
        """set_state_finished"""
        self.state = TaskState.Success
    
    def set_state_imcomplete(self):
        """set_state_imcomplete"""
        self.state = TaskState.Imcomplete
        
    def set_first_flow(self):
        """set_first_flow"""
        self.is_first = True
        
    def is_new(self):
        """get task new"""
        if( self.state != TaskState.Unstart):
            return True
        else :
            return False
        
    def is_ongoing(self):
        """if task is ongoing"""
        if( self.state == TaskState.Ongoing):
            return True
        else :
            return False
    
    def is_finish(self):
        """if task finished"""
        if( self.state == TaskState.Success):
            return True
        else :
            return False
        
    def is_imcomplete(self):
        """is task not finished"""
        if( self.state == TaskState.Imcomplete):
            return True
        else :
            return False
        
    def is_first_flow(self):
        """if 1st flow in flowlist"""
        return self.is_first
    
    def has_fpt(self):
        """do fpt"""
        if( self.fpt_state == FptState.on):
            return True
        else :
            return False
    
    def has_pt(self):
        """do pt"""
        if( self.pt_state == PTState.on):
            return True
        else :
            return False


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File : base.py
@Author : yell
@Desc : feature base 
'''

from settings.setting import Settings
from workspace.path import WorkspacePath
from database.enum import EdaTool, FeatureOption
from flow.flow_db import DbFlow

class FeatureBase():
    """feature base"""
    
    def __init__(self, dir_workspace : str, dir_resource : str = "", eda_tool : EdaTool = EdaTool.IEDA, 
                 feature_option : FeatureOption = FeatureOption.summary, flow : DbFlow = None,
                 input_def : str = None, input_verilog : str = None):
        self.dir_workspace = dir_workspace
        self.dir_resource = dir_resource
        self.eda_tool = eda_tool
        self.feature_option = feature_option
        self.flow = flow
        self.input_def = input_def
        self.input_verilog = input_verilog    
        
        self.workspace = WorkspacePath(dir_workspace)
        
    def generate(self, feature_option : FeatureOption = FeatureOption.NoFeature, reload : bool = False):
        # reset feature option
        if(feature_option != FeatureOption.NoFeature):
            self.feature_option = feature_option
            
        if(self.eda_tool == EdaTool.IEDA):
            return self.generate_db_ieda(reload = reload)
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.generate_db_innovus()
        
    def get_db(self, feature_option : FeatureOption = FeatureOption.NoFeature):
        # reset feature option
        if(feature_option != FeatureOption.NoFeature):
            self.feature_option = feature_option
            
        if(self.eda_tool == EdaTool.IEDA):
            return self.get_db_ieda()
            
        if(self.eda_tool == EdaTool.INNOVUS):
            return self.get_db_innovus()